
/**
 * \file 
 * ines GPIB interface API
 */
#ifndef _GISI_GISU_HXX_
#define _GISI_GISU_HXX_

#include "cplRaw/sm.hxx"
#include "cplRaw/dbg.hxx"
#include "gpib/gumi_user.hxx"
#include "gpibc/gumi_msgs.hxx"
#include <iostream>
#include <vector>

class Block;

namespace INdepGpib100 {
  namespace Umi=INdepUmi100;
  namespace Gif=INdepGpib100;

  typedef int DelE;

  /** A GpibAddress represents a GPIB bus address. A GPIB address has a primary address part
   *  and on optional secondary part.
   */
  class GpibAddress {
    int _iPa;
    bool _fSa;
    int _iSa;
   public:
    /** 
     *  Comparison operator. Two GpibAddress objects are equal if they represent
     *  the same primary and secondary (optional) GPIB address.
     */
    friend bool operator==(const GpibAddress &gaL, const GpibAddress& gaR) {
      return (
        gaL._iPa == gaR._iPa &&
        gaL._fSa == gaR._fSa &&
        gaL._iSa == gaR._iSa 
      );
    }
    friend bool operator!=(const GpibAddress &gaL, const GpibAddress& gaR) { return !operator==(gaL, gaR); }
    bool fSa() const { return _fSa; } 		/**< secondary address present */
    int iPa() const { return _iPa; }		/**< primary address	*/
    int iSa() const { return _iSa; }		/**< secondary address, valid if fSa()  */

    /** GpibAddress constructor. Represents the GPIB address \a iPa and the
     *  optional secondary address \a iSa if \a fSa is true.
     */
    GpibAddress(int iPa, bool fSa=false, int iSa=0): _iPa(iPa), _fSa(fSa), _iSa(iSa) {}

    /** Default Constructor, equivalent to GpibAddress(0).
     */
    GpibAddress(): _iPa(0), _fSa(false), _iSa(0) {}
    ~GpibAddress() {}
  };

  /** An GpibAddrList is a vector of GpibAddress objects.
   */
  class GpibAddrList:public std::vector<GpibAddress> { };
  class GpibStbList:public std::vector<char> { };
  class GpibResList:public std::vector<short> { };

  /** GpibCmds is a list of commands to be send over the GPIB.
   *  It defines several members to construct command sequences easily.
   */

  class GpibCmds {
    enum {
      MAX_CMD=40		/**< Limit of commands to be sent	*/
    };
    char _mpch[MAX_CMD];
    char *_pch;
    GpibAddress _ga;	/* controller address	*/

    bool canput(unsigned i) { return _pch+i < _mpch+MAX_CMD; } 
    void put(int ch) { *(_pch++)=ch; }

   public:

    void setupReceive( const GpibAddress &gaFrom);	/**< Setup to receive from \a gaFrom. inserts UNL MLA [MSA] TAG [SCG] */
    void setupReceive( );				/**< Setup to receive from unspecified talker. Inserts UNL MLA [MSA] */
    void setupSend( const GpibAddress& gaTo );
    void setupSendAdd( const GpibAddress &gaTo );
    void setupSend( const GpibAddrList &gal );
    void cmdAdd( int i);
    void addr( int cmd, const GpibAddress &ga);

    unsigned size() { return _pch-_mpch; }		/**< count of commands buffered so far	*/
    char *mpch() { return _mpch; }    
    char *pch() { return _mpch; }    			/**< buffer to send commands from	*/
    void clear() { _pch=_mpch; }			/**< clear and reuse buffer		*/
    GpibCmds(const GpibAddress &ga):_pch(_mpch), _ga(ga) {} /**< Construct a GpibCmds. \a ga if the controllers (i.e. our) address */
    ~GpibCmds() {}
  };

  /**
   * GPIB user base class (Controller)
   *
   * provides imperative functions for conducting the bus activity.
   *
   * Errors:
   *
   *  Any error in communicating with one of the UMI driver channels
   * results in a BAD condition, i.e. this object is unusable (isBad()).
   *
   *  Errors returned by the UMI driver channel are related to the
   * failed transaction only, i.e. another transaction may be started.
   * (isFail()). 
   * 
   * 
   */
  class GpibControl:public Failable, public SmM, public DebugLog {
    Umi::Umi *_mppum[2];			
    int _iTrans;

    ErrE _eErr;				/* result of last call		*/
    Gif::StatE _fStat;		/* stat of last successful call	*/
    GpibInfo _gi;			/* driver capabilities		*/
    GpibAddress _ga;			/* setup by open()		*/
    int _iPaLast;			/* last foreign primaddr sent	*/
    bool _fSc;				/* system control ?		*/
    bool _fDel;				/* delete _pum with this ?	*/
    const Gpib_gco_Proto _pt;		/* protocol instance		*/
    enum { S_CLOSED, S_OPEN };

    /* messages for GPIB interface */

    Umi::Umi *pum() const { return _mppum[0]; }
    void wait();
    void signal();

    void setTransNext() { _iTrans ++ ; }
    int iTransCur() { return _iTrans; }			/* current transaction id	*/
    typedef INdepGpib100::StatE StatE;

    /*
     * result managers
     */
    GpibControl &setState( GpibResp *pgh ) { 
      _eErr=(ErrE)pgh->eErr(); 
      _fStat=(Gif::StatE)pgh->fStat(); 
      if (_eErr) Failable::setFailed("bus error no", _eErr);
      return *this; 
    }

    GpibControl &setState( GpibResp *pgh, int eState ) { 
      setState(pgh);
      if (isGood()) SmM::setState(eState); 
      return *this; 
    }
    /** UMI failed		*/
    GpibControl &setFailed( Umi::Umi *pum) { 
      Failable::setBad(__FILE__, __LINE__, pum);
      return *this; 
    }	

    GpibControl &setFailed( ErrE eErr ) { 
      if ((_eErr=eErr)) Failable::setFailed("error no", _eErr);
      return *this; 
    }	

  #if 0
    int setFailed( Umi::Umi *pum, int i ) {  return i; }	/* UMI failed		*/
    GpibControl &setFailed( Umi::Umi *pum ) {  return *this; }	/* UMI failed		*/



    //int setFailed( GpibResp *pgh, int i ) { _eErr=(ErrE)pgh->eErr(); return i; }
    //int setOk(int i) { _eErr=INdepGpib100::E_OK; return i; }
    GpibControl &setFailed(ErrE i) { _eErr=i; return *this; }

    void pushReceivedMessages(Umi::UmiControl*, Umi::UmiData*);	/* save the indications and continue */
  #endif


    int cwr(Block &bl, const Umi::Time &ti, const GpibTermOut *gto, GpibTrf *ptr);

    GpibControl &getLines(unsigned *pf, const Umi::Time &tiSettle);

    /** 
     * exchange simple message with standard result check
     */
    GpibControl &xchMsg(const Umi::UmiControl&uc);

   public:
    //friend std::ostream &operator <<( std::ostream &os, const GpibControl &lr );
    // now Failable

    /*
     * state semantics
     * failed state is set by member operations 
     */
    ErrE eErr() const { return _eErr; }
    bool isErr(ErrE e) const { return _eErr == e; }

    //bool isGood() const { return _eErr==0 && pum()->isGood(); }		/* fine 		*/
    //bool isFail() const { return _eErr || !pum()->isGood(); }		/* bad request	 	*/
    //bool isBad() const { return !pum()->isGood(); }			/* bad transport 	*/
    //bool operator!() const { return isFail(); }


    //int init(int iAddr);
    //int fini();


    /** @name Interface Initialization Functions */
    //@{
    /**
     * open the GPIB interface
     * \param ga GPIB address of the interface 
     * \param fSc If true, the interface becomes the system controller
     * After a successful return from open(), the interface is online (i.e. executing the GPIB protocol ).
     */
    GpibControl &open(bool fSc, const GpibAddress &ga);

    /**
     * close the GPIB interface. The interface will be offline (i.e. not executing the GPIB protocol)
     * after a successful return.
     */
    GpibControl &close();
    //@}

    /** @name Low-Level GPIB Control Operations 
     * Lightweight operations that provide complete capabilities at a low abstraction level. 
     */
    //@{

    /** 488.2 11.3.3.4.1 conformant stb generation */
    GpibControl &updateStb(const GpibSta &gst);

    /** send Interface Clear. The interface must be system controller to execute this command */
    GpibControl &ifc(const Umi::Time &ti=Umi::Time(0, 100000));

    /** Set Remote Enable according to f. The interface must be system controller to execute this command. */
    GpibControl &ren(bool f);

    /** 
     * Make the interface the active controller. 
     * \param fSync Take control synchronously if set.
     * The interface must be the controller-in-charge (CIC) in order to execute this command.
     */
    GpibControl &cac(bool fSync);	

    /** 
     * Make the active controller go to standby. 
     * The interface must be the controller-in-charge (CIC) in order to execute this command.
     */
    GpibControl &gts();			/* goto standby */

    GpibControl &ist(bool f);		/**< individual status set */
    GpibControl &iblinec( unsigned *pCap );	/**< capability to sense lines 	*/
    GpibControl &iblines( unsigned *pStat ); 	/**< sense lines 		*/


    GpibControl &rtl();			/**< return to local */
    GpibControl &ppc(int v);		/**< local pp config	 */
    GpibControl &rpp(unsigned char *chPpr, const Umi::Time &ti=Umi::Time());	/**< request parallel ppoll */

    /**
     * wait for any specified flag or event active, failes on timeout
     *
     * \param fStat status bits to wait for
     * \param ti time limit 
     * \param fIgnTmo don't return error on timeout 
     *
     * For any unmasked flag in fWait, wait() returns when the flag is set.
     *
     * For any unmasked event, wait() returns when at least one false-to-true
     * transition has been detected since the most recent call to wait() with
     * that event unmasked.
     *
     * If no mask bit is set, wait() returns after ti has elapsed, 
     * updating the current status word. 
     *
     * if fIgnTmo is set, the E_TMO is not set on timeout.
     *
     */
    GpibControl &wait(StatE fStat, const Umi::Time &ti, bool fIgnTmo=false);	

    /**
     * Return the interface status flags.
     * Any successful operation updates the status flags.
     */
    StatE fStat() { return _fStat; }

    /*
     * sev - synchronize with events
     *
     * returns the interface status and allows to wait for
     * specified events
     *
     */
    //@}

    /** @name Low-Level GPIB I/O Operations 
     * Lightweight operations that provide complete capabilities at a low abstraction level. 
     * All transfer functions return the count actually transfered, even on failure
     */
    //@{

    /**
     * command byte transfer. 
     * \param pch command byte array
     * \param c count of bytes in pch
     * \param ti per byte timeout.
     * \return count of bytes transfered
     *
     * The interface must be the controller-in-charge (CIC) in order to execute this command.
     * Commands are data bytes sent with the GPIB control line ATN true. 
     */
    int cmd( const char *pch, int c, const Umi::Time &ti );

    /**
     * data byte transfer
     * \param pch data byte array
     * \param c count of bytes in pch
     * \param ti per byte timeout.
     * \param gto output termination mode
     * \return count of bytes transfered
     *
     * The interface must be addressed as a talker in order to send data bytes.
     */
    int cwr( const char *pch, int c, const Umi::Time &ti, const GpibTermOut &gto );

    /**
     * data byte transfer, like cwr() but with pch1 and pch2 concatenated
     */
    int cwr( const char *pch1, int c1, const char *pch2, int c2, const Umi::Time &ti, const GpibTermOut & );


    /**
     * conducted read transfer
     *
     * \param pch data byte array
     * \param c maximum count to be read (never more than the count of bytes in pch)
     * \param ti per byte timeout.
     * \param gti input termination mode
     * \return count of bytes transfered
     * \retval peDel if nonzero, set to the input terminator: ' ' = not delimited, 'e' = eos (~EOI) delimited, 'E' = EOI delimited
     *
     * crd() reads at most c bytes.
     * If a device read has been interrupted by a E_CLR failure,
     * E_DCAS is cleared and returned in the status word.
     */
    int crd( char *pch, int c, const Umi::Time &ti, const GpibTermIn &gti, DelE *peDel=0 );

    //@}

    /** @name Configuration and Status Queries
     *
     */
    //@{
    bool isSc() { return _fSc; }		/**< system controller 					*/
    int iPa() { return _ga.iPa(); }		/**< primary GPIB address of this interface		*/
    int iSa() { return _ga.iSa(); }		/**< secondary GPIB address of this interface (if fSa())	*/
    bool fSa() { return _ga.fSa(); }		/**< secondary GPIB addressing enabled for this interface	*/
    const GpibAddress &ga() const { return _ga; } /**< GpibAddress of this interface			*/
    int iPaLast() { return _iPaLast; }		/**< most recent sent foreign primary address		*/
    void clrPaLast(int i) { _iPaLast=i; }	/**< set iPaLast()					*/
    int isOpen() { return isState(S_OPEN); }	/**< true  if the interface is currently opended	*/
    //@}

    /** @name IEEE488.2 conformant functions and protocols.
     *
     * These functions and protocols are specified in the IEEE488.2 standard to provide
     * a vendor-independent set of operations. 
     */
    //@{



    /** IEEE488.2 - 16.2.1 - SEND COMMAND */
    GpibControl &sendCmds(const char *, unsigned *pc, const Umi::Time &ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.2 - SEND SETUP */
    GpibControl &sendSetup(const GpibAddrList &al, const Umi::Time&ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.3 - SEND DATA BYTES */
    GpibControl &sendDataBytes(
      const char *, unsigned *pc, 
      const char *szTerm="\n",
      const GpibTermOut&gto=GpibTermOut().fEnd(true),
      const Umi::Time&ti=Umi::Time(10)
    );

    /** IEEE488.2 - 16.2.4 - SEND */
    GpibControl &send(
      const GpibAddress &, 
      const char *, unsigned *pc, 
      const char *szTerm="\n",
      const GpibTermOut&gto=GpibTermOut().fEnd(true),
      const Umi::Time&ti=Umi::Time(10)
    );

    /** IEEE488.2 - 16.2.4 - SEND (list) */
    GpibControl &send(
      const GpibAddrList &, 
      const char *, unsigned *pc, 
      const char *szTerm="\n",
      const GpibTermOut&gto=GpibTermOut().fEnd(true),
      const Umi::Time&ti=Umi::Time(10)
    );

    /** IEEE488.2 - 16.2.4 - SEND (C string, '\\0' terminated) */
    GpibControl &sendsz(
      const GpibAddress &ga, 
      const char *pch, 
      const char *szTerm="\n",
      const GpibTermOut&gto=GpibTermOut().fEnd(true),
      const Umi::Time&ti=Umi::Time(10)
    ) { 
      unsigned c(strlen(pch));
      return send(ga, pch, &c, szTerm, gto, ti); 
    }

    /** IEEE488.2 - 16.2.5 - RECEIVE SETUP */
    GpibControl &receiveSetup(const GpibAddress &ga, const Umi::Time&ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.6 - RECEIVE RESPONSE MESSAGE */
    GpibControl &rcvRespMsg(char *, unsigned *pc, const GpibTermIn&gti=GpibTermIn(), const Umi::Time&ti=Umi::Time(10), DelE *peDel=0);


    /** IEEE488.2 - 16.2.7 - RECEIVE */
    GpibControl &receive(char *, unsigned *pc, const GpibAddress &, 
      const GpibTermIn&gti=GpibTermIn(),
      const Umi::Time&ti=Umi::Time(10),
      DelE *peDel=0
    );

    /** IEEE488.2 - 16.2.8 - SEND IFC */
    GpibControl &sendIfc(); 

    /** IEEE488.2 - 16.2.9.1 - DEVICE CLEAR SELECTED (list) */
    GpibControl &devClear( const GpibAddrList &al, const Umi::Time&ti=Umi::Time(10));
    /** IEEE488.2 - 16.2.9.1 - DEVICE CLEAR SELECTED */
    GpibControl &clear(const GpibAddress &gad, const Umi::Time &ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.9.2 - DEVICE CLEAR ALL */
    GpibControl &devClear( const Umi::Time&ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.10 - ENABLE LOCAL CONTROLS (list) */
    GpibControl &enableLocal( const GpibAddrList &al, const Umi::Time&ti=Umi::Time(10));
    /** IEEE488.2 - 16.2.10 - ENABLE LOCAL CONTROLS (all) */
    GpibControl &enableLocal( );

    /** IEEE488.2 - 16.2.11 - ENABLE REMOTE */
    GpibControl &enableRemote( const GpibAddrList &al, const Umi::Time&ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.12 - SET RWLS */
    GpibControl &setRWLS( const GpibAddrList &al, const Umi::Time&ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.13 - SEND LLO */
    GpibControl &sendLLO( const Umi::Time&ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.14 - PASS CONTROL */
    GpibControl &passControl( const GpibAddress &al, const Umi::Time&ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.15 - PPOLL */
    GpibControl &ppoll( unsigned *puRes,  const Umi::Time&ti=Umi::Time(0));

    /** IEEE488.2 - 16.2.16 - PARALLEL POLL CONFIGURE */
    GpibControl &ppollConfig( const GpibAddress &ga, int iLine, int iSense, const Umi::Time&ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.17 - PARALLEL POLL UNCONFIGURE (list) */
    GpibControl &ppollUnconfig( const GpibAddrList &ga, const Umi::Time&ti=Umi::Time(10));
    /** IEEE488.2 - 16.2.17 - PARALLEL POLL UNCONFIGURE */
    GpibControl &ppollUnconfig( const Umi::Time&ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.18 - READ STATUS BYTE */
    GpibControl &readStatusByte(char *pch, const GpibAddress &gad, const Umi::Time &ti=Umi::Time(10));

    /** IEEE488.2 - 16.2.19 - TRIGGER */
    GpibControl &trigger(const GpibAddrList &ga, const Umi::Time &ti=Umi::Time(10));

    /** IEEE488.2 - 17.1 - RESET */
    GpibControl &reset(const GpibAddrList &ga, const Umi::Time &ti=Umi::Time(10));

    /** IEEE488.2 - 17.2 - FINDRQS */
    GpibControl &findRqs(char *pch, GpibAddrList::const_iterator *pit, const GpibAddrList &ga, const Umi::Time &ti=Umi::Time(10));


    /** IEEE488.2 - 17.3 - ALLSPOLL */
    GpibControl &allSpoll(GpibStbList *pgs, const GpibAddrList &ga, const Umi::Time &ti=Umi::Time(10));

    /** IEEE488.2 - 17.6 - FINDLSTN */
    GpibControl &findLstn(GpibAddrList *pga, const GpibAddrList &ga, const Umi::Time &tiSettle=Umi::Time(0,1000000), const Umi::Time &ti=Umi::Time(10));

    /** IEEE488.2 - 17.8 - TESTSYS */
    GpibControl &testSys(GpibResList *prl, const GpibAddrList &ga, const Umi::Time &ti=Umi::Time(10));
    //@}

    GpibControl( Umi::Factory *puf);
    GpibControl( Umi::Umi *pum );
    ~GpibControl() ;
  };
  /**  \example k195a.cxx */

  /**
   * construction process wrapper for a GpibControl object
   */
  class GpibInterfaceControl:public Failable {
    GpibControl *_pGuc;
     

   public:
    GpibControl *pGuc() { return _pGuc; } /**< pointer to the GpibControl object created */
    GpibControl *pGucRelease(); /**< pointer to the GpibControl object created, to be deleted by caller */


    /**
     * creates a GpibControl object described by sz.
     * \param sz operating system specific interface descriptor.
     */
    GpibInterfaceControl(const char *sz);
    ~GpibInterfaceControl();
  };


  /** 
   \class GpibTermIn
   \brief GPIB Input Termination Mode Descriptor.

   An object of class GpibTermIn describes the termination mode to be used
   for a GPIB read (input from the GPIB) data transfer.
  */

  /** 
   \class GpibTermOut 
   \brief GPIB Output Termination Mode Descriptor.

   An object of class GpibTermOut describes the termination mode to be used
   for a GPIB write (output to the GPIB) data transfer.
  */

  //@{
  /** 
   \fn GpibTermOut& GpibTermOut::fEnd(bool __fEnd)
   \brief Specifies if the data transfer shall be terminated with
    the END message true. 
    
    \param __fEnd If set, the EOI line is transmitted true
    during the transfer of the last data byte of a data block.
  */

  /** 
   \fn GpibTermOut& GpibTermOut::fEoiEos(bool __fEoiEos)
   \brief Specifies if the EOS byte specified by chEos() shall be sent with
    the END message true. 

    \param __fEoiEos If set, each occurence of the EOS byte is
    sent with the END message true.
  */

  /** 
   \fn GpibTermOut& GpibTermOut::chEos(char __chEos)
   \brief Specifies the EOS byte to be used with fEoiEos().

   chEos() is only used with fEoiEos(true)

    \param __chEos The EOS byte to be used with fEoiEos().
    If fEoiEos() is false, __chEos is ignored.
  */

  /** 
   \fn GpibTermOut& GpibTermOut::fEos7(bool __fEos7)
   \brief Specifies the mode the EOS byte is compared with fEoiEos().

   fEos7() is only used with fEoiEos(true)

    \param __fEos7 If set, bit 7 of chEos() is ignored in comparisions
    of a data byte with the chEos() value. Otherwise all 8 bits are compared.
  */

  /** 
   \fn GpibTermOut& GpibTermOut::fHighSpeed(bool __fHighSpeed)
   \brief Specifies is all data bytes following the first data byte
   shall be transfered in high speed mode.

    \param __fHighSpeed If set, all data bytes following the first data
    byte of a transmission are sent in high-speed mode.
  */

  /** 
   \fn GpibTermOut::GpibTermOut()
   \brief Constructs a default GpibTermOut object. 
   
   The members are assigned to the following defaults: 
   chEos(0), fEnd(false), fEoiEos(false), fEos7(false), fHighSpeed(false)
  */

  /** 
   \fn GpibTermOut::GpibTermOut(char chEos,bool fEnd,bool fEoiEos,bool fEos7,bool fHighSpeed)
   \brief Constructs a fully specified GpibTermOut object. 
  */
  //@}
   


}

namespace INdepUmi100 {
  /** 
   \class Umi 
   \brief communication channel to the hardware driver
  */

  /** 
   \class Factory 
   \brief Factory to create an implementation specific communication channel to the hardware driver
  */
}

  /** \mainpage INES ib++ API for GPIB
   */

#endif
